@@ -36,7 +36,7 @@ import numpy
 import pytess
 import tbdump
 
-VERSION = 1.6
+VERSION = 1.7
 
 CREATE_SUBP = {
     'random': 'create_random_graph',
@@ -73,6 +73,9 @@ EXPORT_SUBP = {
 }
 EXPORT_FORMATS = sorted(EXPORT_SUBP.keys())
 
+EPS = 1e-10
+INFINITY = 2 << 30
+
 MAX_RETRIES = 100
 
 class Graph:
@@ -147,19 +150,34 @@ class Graph:
         return statistics.mean(ratios)
 
     def average_path_length(self):
-        self.floyd_warshall()
         lengths = []
         for u, v in itertools.permutations(self.vertices(), 2):
-            if self.is_reachable(u, v):
-                lengths.append(self.T[u][v])
+            lengths.append(self.shortest_path_length(u, v))
         return statistics.mean(lengths)
 
+    def assortativity(self):
+        """Calcurate the assortativity of a graph.  See Eq. (4) in
+        M. E. J. Newman, `Assortative mixing in networks'."""
+        M = len(self.edges())
+        prod_sum, add_sum, square_sum = 0, 0, 0
+        for e in self.edges():
+            j, k = [self.degree(v) for v in e]
+            prod_sum += j * k
+            add_sum += (j + k) / 2
+            square_sum += (j**2 + k**2) / 2
+        r = prod_sum / M - (add_sum / M)**2
+        r /= square_sum / M - (add_sum / M)**2
+        return r
+
     # vertex ----------------------------------------------------------------
     def vertices(self):
         """Return all vertices in the graph as a list."""
         # FIXME: should implement as a generator
         return self.V.keys()
 
+    def nvertices(self):
+        return len(self.vertices())
+
     def has_vertex(self, v):
         """Check if vertex V exists in the graph."""
         if self.V.get(v, None) is None:
@@ -280,6 +298,9 @@ class Graph:
                     found.append([u, v])
         return found
 
+    def nedges(self):
+        return len(self.edges())
+
     def unique_edges(self):
         """Return all unique edges in the graph as a list.  All multi-edges
         are unified into a single edge."""
@@ -480,7 +501,7 @@ class Graph:
         vertex in the shortest path from source vertex S to vertex 4.  You can
         obtain the shortest path to vertex V by traversing dictionary PREV
         from vertex V back to source node S."""
-        self.expect_directed()
+        self.expect_undirected()
         dist = {}
         prev = {}
         for v in self.vertices():
@@ -489,14 +510,14 @@ class Graph:
 
         S = []
         Q = self.vertices()
-        INFINITY = 2 << 30
         while Q:
             Q = sorted(Q, key=lambda x: dist.get(x, INFINITY))
             u = Q.pop(0)
             if dist.get(u, None) is None:
                 break
             S.append(u)
-            for v in self.successors(u):
+            # NOTE: Use successors() for directeg graph.
+            for v in self.neighbors(u):
                 # FIXME: Must reject multi-edged graph.
                 w = self.get_edge_weight_by_id(u, v, 0) or 1
                 if dist.get(v, None) is None or dist[v] > (
@@ -526,12 +547,20 @@ class Graph:
                     for path in find_path(s, prev):
                         yield path + [t]
 
-        self.expect_directed()
+        # self.expect_directed()
         # Build shortest-path tree if not cached yet.
         if not s in self.P:
             self.dijkstra(s)
         return find_path(s, t)
 
+    def shortest_path_length(self, s, t):
+        if s not in self.T or t not in self.T[s]:
+            self.dijkstra(s)
+        if t in self.T[s]:
+            return self.T[s][t]
+        else:
+            return INFINITY
+
     def dijkstra_all_pairs(self):
         """Compute all-pairs shortest paths using Dijkstra's algorithm."""
         for v in self.vertices():
@@ -539,7 +568,7 @@ class Graph:
 
     def floyd_warshall(self):
         """Compute all-pairs shortest paths using Floyd-Warshall algorithm."""
-        self.expect_directed()
+        # self.expect_directed()
         # Initialize weight matrix.
         path = {}
         next_ = {}
@@ -550,7 +579,6 @@ class Graph:
             path[u][v] = self.get_edge_weight_by_id(u, v, 0) or 1
 
         # Run Floyd-Warshall algorithm to find all-pairs shortest paths.
-        INFINITY = 2 << 30
         for k in self.vertices():
             for u in self.vertices():
                 for v in self.vertices():
@@ -583,7 +611,7 @@ class Graph:
         """Check if all vertices in the graph are mutually connected."""
         v = self.random_vertex()
         explored = self.explore(v)
-        return len(explored) == len(self.vertices())
+        return len(explored) == self.nvertices()
 
     def components(self):
         """Return all components (i.e., connected subgraphs) of the graph.
@@ -606,7 +634,10 @@ class Graph:
         maximal = max(self.components(), key=lambda x: len(x))
         return maximal
 
-    def betweenness(self, v):
+    def degree_centrality(self, v):
+        return self.degree(v) / (self.nvertices() - 1)
+
+    def betweenness_centrality(self, v, normalize=True):
         """Return the betweenness centrality for vertex v.  This program
         implements Algorithm 1 (betweenness centrality in unweighted graphs)
         in U. Brandes, `A Fast Algorithm for Betweeness Centrality,' Journal
@@ -657,7 +688,34 @@ class Graph:
 
         if not v in self.Cb:
             _update_betweenness()
-        return self.Cb[v]
+        if normalize:
+            n = self.nvertices()
+            return self.Cb[v] / ((n - 1) * (n - 2))
+        else:
+            return self.Cb[v]
+
+    def closeness_centrality(self, v):
+        d = sum([self.shortest_path_length(v, u) for u in self.vertices()])
+        n = self.nvertices()
+        return n / d
+
+    def eigenvector_centrality(self, v):
+        adj = self.adjacency_matrix()
+        eigvals, eigvecs = numpy.linalg.eig(adj)
+        # Find the index of the largest eigenvalue.
+        i = max(enumerate(eigvals), key=lambda x: x[1])[0]
+        eigvec = eigvecs[:, i]
+        if eigvec[0] < 0:
+            eigvec = -eigvec
+        return eigvec[v - 1]
+
+    def eccentricity(self, v):
+        lengths = [self.shortest_path_length(v, u) for u in self.vertices()]
+        return max(lengths)
+
+    def eccentricities(self):
+        for v in self.vertices():
+            yield self.eccentricity(v)
 
     # graph ----------------------------------------------------------------
     def copy_graph(self, directed=None):
@@ -701,7 +759,7 @@ class Graph:
     def adjacency_matrix(self):
         """Return the adjacency matrix of the graph as NumPy.ndarray
         object."""
-        N = len(self.vertices())
+        N = self.nvertices()
         m = numpy.zeros((N, N), int)
         for u, v in self.edges():
             m[u - 1, v - 1] += 1
@@ -711,7 +769,7 @@ class Graph:
 
     def diagonal_matrix(self):
         """Return the diagonal matrix of the graph as NumPy.ndarray object."""
-        N = len(self.vertices())
+        N = self.nvertices()
         m = numpy.zeros((N, N), int)
         for v in self.vertices():
             m[v - 1, v - 1] = self.degree(v)
@@ -742,7 +800,7 @@ class Graph:
 
     def natural_connectivity(self):
         """Return the natural connectivity from spectral graph theory."""
-        N = len(self.vertices())
+        N = self.nvertices()
         lmbda = self.adjacency_matrix_eigvals()
         return math.log(sum(numpy.exp(lmbda)) / N)
 
@@ -753,13 +811,13 @@ class Graph:
 
     def effective_resistance(self):
         """Return the effective resistance from spectral graph theory."""
-        N = len(self.vertices())
+        N = self.nvertices()
         mu = self.laplacian_matrix_eigvals()
         return N * sum([1 / (mu + 1e-100) for mu in mu[1:]])
 
     def spanning_tree_count(self):
         """Return the spanning tree count from spectral graph theory."""
-        N = len(self.vertices())
+        N = self.nvertices()
         mu = self.laplacian_matrix_eigvals()
         return functools.reduce(lambda x, y: x * y,
                                 [1 / (mu + 1e-100) for mu in mu[1:]]) / N
@@ -768,8 +826,8 @@ class Graph:
     def header_string(self, comment='# '):
         date = time.strftime('%Y/%M/%D %H:%M:%S', time.localtime())
         atype = 'directed' if self.is_directed() else 'undirected'
-        vcount = len(self.vertices())
-        ecount = len(self.edges())
+        vcount = self.nvertices()
+        ecount = self.nedges()
         astr = f"{comment}Generated by graph-tools (version {VERSION}) at {date}]\n{comment}{atype}, {vcount} vertices, {ecount} edges\n"
         return astr
 
@@ -785,21 +843,19 @@ class Graph:
             return None
         return method(*args, **kwargs)
 
-    def create_random_graph(self, N=10, E=20, no_multiedge=False):
+    def create_random_graph(self, N=10, E=20, multiedged=False):
         """Create an instance of *connected* random graphs with N vertices and
-        E edges.  A generated graph might be multiedged.  You can specify
-        NO_MULTIEDGE to force a single-edged graph."""
+        E edges.  A generated graph is non-multiedged.  You can specify
+        MULTIEDGED to allow a multiedged graph."""
         if E < N:
             die('create_random_graph: too small number of edges')
-
-        for v in range(1, N + 1):
-            self.add_vertex(v)
+        self.add_vertices(*range(1, N + 1))
 
         # Add first (N - 1) edges for making sure connectivity.
         for i in range(1, N):
             u = i + 1
             v = random.randrange(1, u)
-            if random.uniform(0, 1) >= .5:
+            if random.random() <= .5:
                 self.add_edge(u, v)
             else:
                 self.add_edge(v, u)
@@ -809,12 +865,15 @@ class Graph:
             # FIXME: Avoid cycle edges, but this may take log time.
             ntries = 1
             while ntries < MAX_RETRIES:
-                u = random.randrange(1, N + 1)
-                v = random.randrange(1, N + 1)
-                if not no_multiedge and u != v:
-                    break
-                if no_multiedge and u != v and not self.has_edge(u, v):
+                u = self.random_vertex()
+                v = self.random_vertex()
+                if u == v:
+                    continue
+                if multiedged:
                     break
+                else:
+                    if not self.has_edge(u, v):
+                        break
             self.add_edge(u, v)
         return self
 
@@ -955,7 +1014,7 @@ class Graph:
 
     # Tree BA graph.
     def create_treeba_graph(self, N=10, alpha=1):
-        self.expect_directed()
+        # self.expect_directed()
         attract = [0] * N
         # Create an initial vertex.
         v = 1
@@ -965,16 +1024,10 @@ class Graph:
         # Create a vertex and attach to another using preferential attachment.
         for u in range(2, N + 1):
             # Randomly choose a vertex with a probability proportional to attract.
-            total = sum(attract)
-            chosen = random.uniform(0, total)
-            accum = 0
-            for v in range(1, N + 1):
-                accum += attract[v - 1]
-                if chosen < accum:
-                    self.add_edge(u, v)
-                    attract[u - 1] = alpha + len(self.in_edges(u))
-                    attract[v - 1] = alpha + len(self.in_edges(v))
-                    break
+            v = random.choices(range(1, N + 1), weights=attract, k=1)[0]
+            self.add_edge(u, v)
+            attract[u - 1] = alpha + len(self.in_edges(u))
+            attract[v - 1] = alpha + len(self.in_edges(v))
         return self
 
     def create_generalized_barabasi_graph(self, N=10, m=2, gamma=3, m0=2):
@@ -994,8 +1047,8 @@ class Graph:
             self.add_vertex(u)
 
             # Attach to a vertex using preferential attachment.
-            vcount = len(self.vertices()) - 1
-            ecount = len(self.edges())
+            vcount = self.nvertices() - 1
+            ecount = self.nedges()
             for j in range(1, m + 1):
                 # NOTE: Preferential-attachement with probability A + in_degree.
                 total = A * vcount + ecount
